#include "2d_physics_environment.h"

PhysicsEnvironment2D::PhysicsEnvironment2D()
{
	
}

PhysicsEnvironment2D::PhysicsEnvironment2D(const PhysicsEnvironment2D& other)
{
}

PhysicsEnvironment2D::~PhysicsEnvironment2D()
{
}

bool PhysicsEnvironment2D::Initialize(abfw::Vector2 gravity)
{
	_gravity = gravity;

	return true;
}

void PhysicsEnvironment2D::Shutdown()
{
	while (_objectList.size())
	{
		_objectList.pop_back();
	}
}

void PhysicsEnvironment2D::StepSimulation(float time)
{
	for (list<PhysicsObject2D*>::const_iterator o = _objectList.begin(); o != _objectList.end(); o++)
	{
		//If the object has no mass then it is a stationary object so movement physics do not need to be calculated
		if ((*o)->GetRigidbody()->GetMass() > 0.0f)
		{
			DirectForce netForce( 0, 0 );

			//Calculate the gravitational force on this object
			DirectForce gravitationalForce( ((*o)->GetRigidbody()->GetMass()*(_gravity.x * GRAVITATIONAL_CONSTANT)), ((*o)->GetRigidbody()->GetMass()*(_gravity.y * GRAVITATIONAL_CONSTANT)) );
			(*o)->GetRigidbody()->ApplyForce(gravitationalForce);

			if ((*o)->GetRigidbody()->GetNumForces() > 1)
			{
				CalculateResultantForce(netForce, (*o)->GetRigidbody()->ReturnForce(), (*o)->GetRigidbody()->ReturnForce());
				
				//Apply the remaining forces to the object
				while ((*o)->GetRigidbody()->GetNumForces() > 0)
				{
					CalculateResultantForce(netForce, netForce, (*o)->GetRigidbody()->ReturnForce());
				}
			}
			else
			{
				netForce = (*o)->GetRigidbody()->ReturnForce();
			}

			//Calculate the acceleration of the object
			abfw::Vector2 objAccel( (netForce.x_comp / (*o)->GetRigidbody()->GetMass()), (netForce.y_comp / (*o)->GetRigidbody()->GetMass()) );
			(*o)->GetRigidbody()->SetAcceleration(objAccel);

			//Calculate the new speed of the object
			abfw::Vector2 objVelocity( ((*o)->GetRigidbody()->GetVelocity().x + (objAccel.x * time)), ((*o)->GetRigidbody()->GetVelocity().y + (objAccel.y * time)) );
			(*o)->GetRigidbody()->SetVelocity(objVelocity);

			//Calculate the initial potential position of the object
			abfw::Vector2 objPosition( ((*o)->GetPosition().x + (objVelocity.x * time)), ((*o)->GetPosition().y + (objVelocity.y * time)) );
			(*o)->SetPotentialPosition(objPosition);
		}
		else
		{
			(*o)->SetPotentialPosition((*o)->GetPosition());
		}
	}

	//Iterate through the object list and check for collisions
	for (list<PhysicsObject2D*>::iterator physObject1 = _objectList.begin(); physObject1 != _objectList.end(); physObject1++)
	{
		for (list<PhysicsObject2D*>::iterator physObject2 = _objectList.begin(); physObject2 != _objectList.end(); physObject2++)
		{
			if ((*physObject1) != (*physObject2))
			{
				CheckCollision(*(*physObject1), *(*physObject2));
			}
		}
		(*physObject1)->SetPosition((*physObject1)->GetPotentialPosition());
	}
}

void PhysicsEnvironment2D::AddPhysicsObject(PhysicsObject2D* physObject)
{
	if (physObject)
	{
		_objectList.push_back(physObject);
	}
}

void PhysicsEnvironment2D::CheckCollision(PhysicsObject2D& physObject1, PhysicsObject2D& physObject2)
{
	//Decide which collision detection system to use based on collider types
	if (physObject1.GetCollider()->GetColliderTypeIndex() == 0 && physObject2.GetCollider()->GetColliderTypeIndex() == 0)
	{
		BoxToBoxCollision(physObject1, physObject2);
	}
	else if (physObject1.GetCollider()->GetColliderTypeIndex() == 0 && physObject2.GetCollider()->GetColliderTypeIndex() == 1)
	{
		BoxToCircleCollision(physObject1, physObject2);
	}
	else if (physObject1.GetCollider()->GetColliderTypeIndex() == 1 && physObject2.GetCollider()->GetColliderTypeIndex() == 0)
	{
		BoxToCircleCollision(physObject2, physObject1);
	}
	else if (physObject1.GetCollider()->GetColliderTypeIndex() == 1 && physObject2.GetCollider()->GetColliderTypeIndex() == 1)
	{
		CircleToCircleCollision(physObject1, physObject2);
	}
}

bool PhysicsEnvironment2D::BoxToBoxCollision(PhysicsObject2D& physObject1, PhysicsObject2D& physObject2)
{
	int halfWidth1, halfHeight1;
	int halfWidth2, halfHeight2;
	static_cast<BoundingBox*>(physObject1.GetCollider())->GetDimensions(halfWidth1, halfHeight1);
	static_cast<BoundingBox*>(physObject2.GetCollider())->GetDimensions(halfWidth2, halfHeight2);

	halfWidth1 *= 0.5f;
	halfHeight1 *= 0.5f;

	halfWidth2 *= 0.5f;
	halfHeight2 *= 0.5f;

	abfw::Vector2 pos1 = physObject1.GetPotentialPosition();
	abfw::Vector2 pos2 = physObject2.GetPotentialPosition();

	//int potentialXs[2] = { pos1.x, pos2.x };

	//clockwise starting top
	int obj1Sides[4] = { pos1.y + halfHeight1, pos1.x + halfWidth1, pos1.y - halfHeight1, pos1.x - halfWidth1 };
	int obj2Sides[4] = { pos2.y + halfHeight2, pos2.x + halfWidth2, pos2.y - halfHeight2, pos2.x - halfWidth2 };

	//Calculate where the collisions have occurred and adjust positions of the two objects accordingly
	//Calculate X axis collisions
	if (((physObject1.GetPosition().y + halfHeight1 > obj2Sides[2]) && (physObject1.GetPosition().y - halfHeight1 < obj2Sides[0])) && physObject1.GetRigidbody()->GetVelocity().x !=0)
	{
		if (((obj1Sides[1]) > (obj2Sides[3])) && ((obj1Sides[3]) < (obj2Sides[3])))
		{
			if (physObject1.GetRigidbody()->GetVelocity().x != 0)
			{
				if (physObject2.GetRigidbody()->GetVelocity().x != 0)
				{
					int diff = (obj1Sides[1] - obj2Sides[3]) / 2;
					physObject1.SetPotentialPosition(abfw::Vector2((pos1.x - diff), pos1.y));
					physObject2.SetPotentialPosition(abfw::Vector2((pos2.x + diff), pos2.y));
					physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject1.GetRigidbody()->GetVelocity().y));
					physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject2.GetRigidbody()->GetVelocity().y));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::LEFT);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::RIGHT);
					return true;
				}
				else
				{
					physObject1.SetPotentialPosition(abfw::Vector2((obj2Sides[3] - halfWidth1), pos1.y));
					physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject1.GetRigidbody()->GetVelocity().y));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::LEFT);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::RIGHT);
					return true;
				}
			}
			else
			{
				if (physObject2.GetRigidbody()->GetVelocity().x != 0)
				{
					physObject2.SetPotentialPosition(abfw::Vector2((obj1Sides[1] + halfWidth2), pos2.y));
					physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject2.GetRigidbody()->GetVelocity().y));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::LEFT);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::RIGHT);
					return true;
				}
			}
		}
		else if ((obj1Sides[3]<obj2Sides[1]) && (obj1Sides[1] > obj2Sides[1]))
		{
			if (physObject1.GetRigidbody()->GetVelocity().x != 0)
			{
				if (physObject2.GetRigidbody()->GetVelocity().x != 0)
				{
					int diff = (obj2Sides[1] - obj1Sides[3]) / 2;
					physObject1.SetPotentialPosition(abfw::Vector2((pos1.x + diff), pos1.y));
					physObject2.SetPotentialPosition(abfw::Vector2((pos2.x - diff), pos2.y));
					physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject1.GetRigidbody()->GetVelocity().y));
					physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject2.GetRigidbody()->GetVelocity().y));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::RIGHT);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::LEFT);
					return true;
				}
				else
				{
					physObject1.SetPotentialPosition(abfw::Vector2((obj2Sides[1] + halfWidth1), pos1.y));
					physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject1.GetRigidbody()->GetVelocity().y));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::RIGHT);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::LEFT);
					return true;
				}
			}
			else
			{
				if (physObject2.GetRigidbody()->GetVelocity().x != 0)
				{
					physObject2.SetPotentialPosition(abfw::Vector2((obj1Sides[3] - halfWidth2), pos2.y));
					physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject2.GetRigidbody()->GetVelocity().y));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::RIGHT);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::LEFT);
					return true;
				}
			}
		}
	}

	//Calculate Y axis collision
	if (((physObject1.GetPosition().x + halfWidth1 > obj2Sides[3]) && (physObject1.GetPosition().x - halfWidth1 < obj2Sides[1])) && physObject1.GetRigidbody()->GetVelocity().y != 0)
	{
		if (((obj1Sides[2]) < (obj2Sides[0])) && ((obj1Sides[0]) > (obj2Sides[0])))
		{
			if (physObject1.GetRigidbody()->GetVelocity().y != 0)
			{
				if (physObject2.GetRigidbody()->GetVelocity().y != 0)
				{
					int diff = (obj1Sides[0] - obj2Sides[0]) / 2;
					physObject1.SetPotentialPosition(abfw::Vector2((physObject1.GetPotentialPosition().x), (pos1.y - diff)));
					physObject2.SetPotentialPosition(abfw::Vector2((physObject2.GetPotentialPosition().x), (pos2.y + diff)));
					physObject1.GetRigidbody()->SetVelocity(abfw::Vector2((physObject1.GetRigidbody()->GetVelocity().x*0.9), 0));
					physObject2.GetRigidbody()->SetVelocity(abfw::Vector2((physObject2.GetRigidbody()->GetVelocity().x*0.9), 0));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::TOP);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::BOTTOM);
					return true;
				}
				else
				{
					physObject1.SetPotentialPosition(abfw::Vector2((physObject1.GetPotentialPosition().x), (obj2Sides[0] + halfHeight1)));
					physObject1.GetRigidbody()->SetVelocity(abfw::Vector2((physObject1.GetRigidbody()->GetVelocity().x*0.9), 0));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::TOP);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::BOTTOM);
					return true;
				}
			}
			else
			{
				if (physObject2.GetRigidbody()->GetVelocity().y != 0)
				{
					physObject2.SetPotentialPosition(abfw::Vector2((physObject2.GetPotentialPosition().x), (obj1Sides[2] - halfHeight2)));
					physObject2.GetRigidbody()->SetVelocity(abfw::Vector2((physObject2.GetRigidbody()->GetVelocity().x*0.9), 0));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::TOP);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::BOTTOM);
					return true;
				}
			}
		}
		else if ((obj1Sides[0]>obj2Sides[2]) && (obj1Sides[2] < obj2Sides[2]))
		{
			if (physObject1.GetRigidbody()->GetVelocity().y != 0)
			{
				if (physObject2.GetRigidbody()->GetVelocity().y != 0)
				{
					int diff = (obj2Sides[2] - obj1Sides[0]) / 2;
					physObject1.SetPotentialPosition(abfw::Vector2((physObject1.GetPotentialPosition().x), (pos1.y + diff)));
					physObject2.SetPotentialPosition(abfw::Vector2((physObject2.GetPotentialPosition().x), (pos2.y - diff)));
					physObject1.GetRigidbody()->SetVelocity(abfw::Vector2((physObject1.GetRigidbody()->GetVelocity().x*0.9), 0));
					physObject2.GetRigidbody()->SetVelocity(abfw::Vector2((physObject2.GetRigidbody()->GetVelocity().x*0.9), 0));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::BOTTOM);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::TOP);
					return true;
				}
				else
				{
					physObject1.SetPotentialPosition(abfw::Vector2((physObject1.GetPotentialPosition().x), (obj2Sides[2] - halfHeight1)));
					physObject1.GetRigidbody()->SetVelocity(abfw::Vector2((physObject1.GetRigidbody()->GetVelocity().x*0.9), 0));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::BOTTOM);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::TOP);
					return true;
				}
			}
			else
			{
				if (physObject2.GetRigidbody()->GetVelocity().y != 0)
				{
					physObject2.SetPotentialPosition(abfw::Vector2((physObject2.GetPotentialPosition().x), (obj1Sides[0] + halfHeight2)));
					physObject2.GetRigidbody()->SetVelocity(abfw::Vector2((physObject2.GetRigidbody()->GetVelocity().x*0.9), 0));
					physObject1.PushCollisionData(physObject2.GetObjectTag(), collision_position::BOTTOM);
					physObject2.PushCollisionData(physObject1.GetObjectTag(), collision_position::TOP);
					return true;
				}
			}
		}
	}

	return false;
}

bool PhysicsEnvironment2D::BoxToCircleCollision(PhysicsObject2D& physObject1, PhysicsObject2D& physObject2)
{
	int halfWidth, halfHeight;
	int radius;
	float smallestDistance = -1;
	int collisionPoint = -1;

	radius = static_cast<BoundingCircle*>(physObject2.GetCollider())->GetRadius();
	static_cast<BoundingBox*>(physObject1.GetCollider())->GetDimensions(halfWidth, halfHeight);

	halfWidth *= 0.5f;
	halfHeight *= 0.5f;

	abfw::Vector2 position = physObject1.GetPotentialPosition();
	abfw::Vector2 position2 = physObject2.GetPotentialPosition();

	//Make an array of 'focus points' which will be checked for collisions with the circle
	//Currently manual but there could be potential for inputting a precision level and automatically generating an array of points to check
	abfw::Vector2 points[8] =
	{
		abfw::Vector2(position.x - halfWidth, position.y + halfHeight),
		abfw::Vector2(position.x, position.y - halfHeight),
		abfw::Vector2(position.x + halfWidth, position.y + halfHeight),
		abfw::Vector2(position.x + halfWidth, position.y),
		abfw::Vector2(position.x + halfWidth, position.y - halfHeight),
		abfw::Vector2(position.x, position.y - halfHeight),
		abfw::Vector2(position.x - halfWidth, position.y - halfHeight),
		abfw::Vector2(position.x - halfWidth, position.y)
	};

	//Check if any points have collisions with the circle and if so which is closest to the centre
	for (int i = 0; i < 8; i++)
	{
		float distance = sqrt(pow((position2.x - position.x), 2) + pow((position2.y - position.y), 2));

		if (distance < radius)
		{
			if (smallestDistance == -1 || distance < smallestDistance)
			{
				collisionPoint = i;
				smallestDistance = distance;
			}
		}
	}

	//If there is a collision move the square to the point of collision
	if (collisionPoint != -1)
	{
		float angle;
		abfw::Vector2 vecDistance( (position2.x - position.x), (position2.y - position.y) );
		angle = CalculateAngleInCircle(vecDistance.x, vecDistance.y);
		abfw::Vector2 transformPoint;

		if (angle == 0)
		{
			transformPoint.x = position2.x;
			transformPoint.y = position2.y + radius;
		}
		else if (angle == 90)
		{
			transformPoint.x = position2.x + radius;
			transformPoint.y = position2.y;
		}
		else if (angle == 180)
		{
			transformPoint.x = position2.x;
			transformPoint.y = position2.y - radius;
		}
		else if (angle == 270)
		{
			transformPoint.x = position2.x - radius;
			transformPoint.y = position2.y;
		}
		else
		{
			transformPoint.x = (cos(angle*PI / 180) * radius);
			transformPoint.y = (sin(angle*PI / 180) * radius);
		}

		switch (collisionPoint)
		{
		case 0:
			physObject1.SetPotentialPosition(abfw::Vector2( (transformPoint.x + halfWidth), (transformPoint.y - halfHeight) ));
			physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(0, 0));
			physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(0, 0));
			break;
		case 1:
			physObject1.SetPotentialPosition(abfw::Vector2( (transformPoint.x), (transformPoint.y - halfHeight) ));
			physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(physObject1.GetRigidbody()->GetVelocity().x, 0));
			physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(physObject2.GetRigidbody()->GetVelocity().x, 0));
			break;
		case 2:
			physObject1.SetPotentialPosition(abfw::Vector2( (transformPoint.x - halfWidth), (transformPoint.y - halfHeight) ));
			physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(0, 0));
			physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(0, 0));
			break;
		case 3:
			physObject1.SetPotentialPosition(abfw::Vector2( (transformPoint.x - halfWidth), (transformPoint.y) ));
			physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject1.GetRigidbody()->GetVelocity().y));
			physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject2.GetRigidbody()->GetVelocity().y));
			break;
		case 4:
			physObject1.SetPotentialPosition(abfw::Vector2( (transformPoint.x - halfWidth), (transformPoint.y + halfHeight) ));
			physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(0, 0));
			physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(0, 0));
			break;
		case 5:
			physObject1.SetPotentialPosition(abfw::Vector2( (transformPoint.x), (transformPoint.y + halfHeight) ));
			physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(physObject1.GetRigidbody()->GetVelocity().x, 0));
			physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(physObject2.GetRigidbody()->GetVelocity().x, 0));
			break;
		case 6:
			physObject1.SetPotentialPosition(abfw::Vector2( (transformPoint.x + halfWidth), (transformPoint.y + halfHeight) ));
			physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(0, 0));
			physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(0, 0));
			break;
		case 7:
			physObject1.SetPotentialPosition(abfw::Vector2( (transformPoint.x + halfWidth), (transformPoint.y) ));
			physObject1.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject1.GetRigidbody()->GetVelocity().y));
			physObject2.GetRigidbody()->SetVelocity(abfw::Vector2(0, physObject2.GetRigidbody()->GetVelocity().y));
			break;
		}

		return true;
	}

	return false;
}

bool PhysicsEnvironment2D::CircleToCircleCollision(PhysicsObject2D& physObject1, PhysicsObject2D& physObject2)
{
	abfw::Vector2 position = physObject1.GetPotentialPosition();
	abfw::Vector2 position2 = physObject2.GetPotentialPosition();
	abfw::Vector2 vecDistance( (position2.x-position.x), (position2.y-position.y) );

	float distance = sqrt(pow((position2.x - position.x), 2) + pow((position2.y - position.y), 2));

	//Get the radius of each circle collider
	float radius1 = static_cast<BoundingCircle*>(physObject1.GetCollider())->GetRadius();
	float radius2 = static_cast<BoundingCircle*>(physObject2.GetCollider())->GetRadius();

	if (distance > (radius1 + radius2))
	{
		float angle = CalculateAngleInCircle(vecDistance.x, vecDistance.y);

		abfw::Vector2 transformPoint;

		if (angle == 0)
		{
			transformPoint.x = position2.x;
			transformPoint.y = position2.y + radius2;
		}
		else if (angle == 90)
		{
			transformPoint.x = position2.x + radius2;
			transformPoint.y = position2.y;
		}
		else if (angle == 180)
		{
			transformPoint.x = position2.x;
			transformPoint.y = position2.y - radius2;
		}
		else if (angle == 270)
		{
			transformPoint.x = position2.x - radius2;
			transformPoint.y = position2.y;
		}
		else
		{
			transformPoint.x = (cos(angle*PI / 180) * radius2);
			transformPoint.y = (sin(angle*PI / 180) * radius2);
		}

		return true;
	}

	return false;
}