#include "sprite_application.h"
#include <system/platform.h>
#include <iostream>

SpriteApplication::SpriteApplication(abfw::Platform& platform) :
	abfw::Application(platform),
	sprite_renderer_(NULL)
{
		scene_ = NULL;
		input_ = NULL;
}

SpriteApplication::~SpriteApplication()
{
}

void SpriteApplication::Init()
{
	game_mode_ = mode::TITLE_SCENE;

	// load the font to draw the on-screen text
	bool font_loaded = font_.Load("comic_sans", platform_);
	if(!font_loaded)
	{
		std::cout << "Font failed to load." << std::endl;
		exit(-1);
	}

	// create a sprite renderer to draw the sprites
	sprite_renderer_ = platform_.CreateSpriteRenderer();

	// create a controller manager to receive controller input
	input_ = new InputClass;
	
	input_->Initialize(platform_);
	
	scene_creator_.Initialize(platform_, input_, NULL);
	scene_selector_.Initialize(platform_, input_, NULL);
	title_scene_.Initialize(platform_, input_, NULL);

	switch (game_mode_)
	{
	case(mode::SCENE) :
		scene_ = new SceneClass;
		scene_->Initialize(platform_, input_, "003.txt");
		break;
	case(mode::SCENE_CREATION):
		break;
	case(mode::SCENE_SELECTION) :
		break;
	case(mode::TITLE_SCENE) :
		break;
	}
	
	//Load in the level list so that the correct scenes can be loaded
	LoadLevelList();
}

void SpriteApplication::CleanUp()
{
	// free up controller manager
	input_->Shutdown();
	delete input_;
	input_ = NULL;

	// free up the sprite renderer
	delete sprite_renderer_;
	sprite_renderer_ = NULL;

	//shutdown the necessary components
	scene_creator_.Shutdown();
	scene_selector_.Shutdown();
	title_scene_.Shutdown();
}

bool SpriteApplication::Update(float ticks)
{
	int result = 0; 
	// calculate the frame rate
	frame_rate_ = 1.0f / ticks;

	input_->Frame();

	//update in the current mode
	switch (game_mode_)
	{
	case(mode::SCENE_CREATION) :
		result = scene_creator_.Frame(ticks);
		if (result)
		{
			//Update the stored level list to new version
			scene_creator_.LoadLevelList();
			scene_selector_.LoadLevelCount();
			LoadLevelList();
			scene_creator_.ClearScene();

			game_mode_ = mode::SCENE_SELECTION;
		}
		break;
	case(mode::SCENE) :
		result =scene_->Frame(ticks);
		if (result)
		{
			delete scene_;
			scene_ = NULL;
			game_mode_ = mode::SCENE_SELECTION;
		}
		break;
	case(mode::SCENE_SELECTION) :
		result = scene_selector_.Frame(ticks);
		if (result == -1)
		{
			game_mode_ = mode::SCENE_CREATION;
		}
		else if (result != 0)
		{
			int level = result - 1;

			scene_ = new SceneClass;
			scene_->Initialize(platform_, input_, filenames_[level].c_str());

			game_mode_ = mode::SCENE;
		}

		break;

	case(mode::TITLE_SCENE) :
		result = title_scene_.Frame(ticks);
		if (result)
		{
			game_mode_ = mode::SCENE_SELECTION;
		}
	}

	// return if the application is still running
	// this should always be true on PS Vita
	return true;
}

void SpriteApplication::Render()
{
	// set up sprite renderer for drawing
	sprite_renderer_->Begin();

	//render the appropriate game screen based on mode
	switch (game_mode_)
	{
	case (mode::SCENE_CREATION) :

		scene_creator_.Render((abfw::SpriteRendererVita*)sprite_renderer_, font_);

		break;

	case(mode::SCENE) :

		scene_->Render((abfw::SpriteRendererVita*)sprite_renderer_, font_);

		break;

	case(mode::SCENE_SELECTION) :
		scene_selector_.Render((abfw::SpriteRendererVita*)sprite_renderer_, font_);
		break;

	case(mode::TITLE_SCENE) :
		title_scene_.Render((abfw::SpriteRendererVita*)sprite_renderer_, font_);
		break;
	}


	// tell sprite renderer that all sprites have been drawn
	sprite_renderer_->End();
}

bool SpriteApplication::LoadLevelList()
{
	int bufferPosition;
	int tempBufferPosition;
	char input;
	int stringLength;
	char* charInput;
	int i;
	int hundreds, tens, ones;
	File* level_file = platform_.CreateFile();

	bool success = level_file->Open("level_list.txt");
	if (success)
	{
		while (filenames_.size()>0)
		{
			filenames_.pop_back();
		}

		char* buffer = NULL;
		Int32 file_size;
		success = level_file->GetSize(file_size);
		file_size++;
		if (success)
		{
			buffer = new char[file_size];
			Int32 bytes_read;
			success = level_file->Read(buffer, file_size, bytes_read);

			if (success)
				level_file->Close();

			if (success)
			{
				//Read the level count
				bufferPosition = 0;
				hundreds = buffer[bufferPosition] - '0';
				bufferPosition++;
				tens = buffer[bufferPosition] - '0';
				bufferPosition++;
				ones = buffer[bufferPosition] - '0';

				level_count_ = (hundreds * 100) + (tens * 10) + ones;

				//Read in all the filenames
				for (int x = 0; x < level_count_; x++)
				{
					bufferPosition++;
					//Move to the filename
					input = buffer[bufferPosition];
					while (input < 48)
					{
						bufferPosition++;
						input = buffer[bufferPosition];
					}

					//Calculate the length of the filename
					tempBufferPosition = bufferPosition;
					stringLength = 0;
					while (input  > 47)
					{
						tempBufferPosition++;
						stringLength++;
						input = buffer[tempBufferPosition];
						if (input == '.')
						{
							stringLength++;
							tempBufferPosition++;
							input = buffer[tempBufferPosition];
						}
					}
					//Read in the filenames
					charInput = new char[stringLength + 1];
					i = 0;
					input = buffer[bufferPosition];
					while (input >47)
					{
						charInput[i] = input;
						i++;
						bufferPosition++;
						input = buffer[bufferPosition];
						if (input == '.')
						{
							charInput[i] = input;
							bufferPosition++;
							i++;
							input = buffer[bufferPosition];
						}
					}
					//Add a string terminator to the end of each file
					charInput[stringLength] = '\0';
					string stringInput(charInput);
					filenames_.push_back(stringInput);
				}

				return true;
			}
		}
	}

	return false;
}